import pickle
import errno
import re
import os
import ipaddress
from pickle import UnpicklingError
from tempfile import mkstemp
from time import sleep
import logging
import signal
from typing import List, Tuple, Optional, Type, TYPE_CHECKING

from blueman.Constants import DHCP_CONFIG_FILE
from blueman.Functions import have
from _blueman import create_bridge, destroy_bridge, BridgeException
from subprocess import call, Popen, PIPE

if TYPE_CHECKING:
    from typing_extensions import Protocol

    class DHCPHandler(Protocol):
        def __init__(self, netconf: "NetConf"):
            ...

        def do_apply(self) -> None:
            ...

        def do_remove(self) -> None:
            ...


class NetworkSetupError(Exception):
    pass


def is_running(name: str, pid: int) -> bool:
    if not os.path.exists(f"/proc/{pid}"):
        return False

    with open(f"/proc/{pid}/cmdline") as f:
        return name in f.readline().replace("\0", " ")


def kill(pid: Optional[int], name: str) -> bool:
    if pid and is_running(name, pid):
        print('Terminating ' + name)
        os.kill(pid, signal.SIGTERM)
        return True
    return False


def read_pid_file(fname: str) -> Optional[int]:
    try:
        with open(fname) as f:
            return int(f.read())
    except (OSError, ValueError):
        return None


def get_dns_servers() -> str:
    dns_servers = ''
    with open("/etc/resolv.conf") as f:
        for line in f:
            match = re.search(r"^nameserver ((?:[0-9]{1,3}\.){3}[0-9]{1,3}$)", line)
            if match:
                dns_servers += f"{match.group(1)}, "

        dns_servers = dns_servers.strip(", ")

    return dns_servers


def get_binary(*names: str) -> str:
    for name in names:
        path = have(name)
        if path:
            return path
    raise FileNotFoundError(f"{' '.join(names)} not found")


class DnsMasqHandler:
    def __init__(self, netconf: "NetConf") -> None:
        self.pid: Optional[int] = None
        self.netconf = netconf

    def do_apply(self) -> None:
        if not self.netconf.locked("dhcp") or self.netconf.ip4_changed:
            if self.netconf.ip4_changed:
                self.do_remove()

            assert self.netconf.ip4_address is not None

            ipiface = ipaddress.ip_interface('/'.join((self.netconf.ip4_address, '255.255.255.0')))
            cmd = [get_binary("dnsmasq"), "--port=0", "--pid-file=/var/run/dnsmasq.pan1.pid", "--except-interface=lo",
                   "--interface=pan1", "--bind-interfaces",
                   f"--dhcp-range={ipiface.network[2]},{ipiface.network[-2]},60m",
                   f"--dhcp-option=option:router,{self.netconf.ip4_address}"]

            logging.info(cmd)
            p = Popen(cmd, stderr=PIPE)

            error = p.communicate()[1]

            if not error:
                logging.info("Dnsmasq started correctly")
                with open("/var/run/dnsmasq.pan1.pid") as f:
                    self.pid = int(f.read())
                logging.info(f"pid {self.pid}")
                self.netconf.lock("dhcp")
            else:
                error_msg = error.decode("UTF-8").strip()
                logging.info(error_msg)
                raise NetworkSetupError(f"dnsmasq failed to start: {error_msg}")

    def do_remove(self) -> None:
        if self.netconf.locked("dhcp"):
            if not self.pid:
                pid = read_pid_file("/var/run/dnsmasq.pan1.pid")
            else:
                pid = self.pid

            if not kill(pid, 'dnsmasq'):
                logging.info("Stale dhcp lockfile found")
            self.netconf.unlock("dhcp")


DHCPDSUBNET = '''#### BLUEMAN AUTOMAGIC SUBNET ####
# Everything inside this section is destroyed after config change
subnet %(ip_mask)s netmask %(netmask)s {
  option domain-name-servers %(dns)s;
  option subnet-mask %(netmask)s;
  option routers %(rtr)s;
  range %(start)s %(end)s;
}
#### END BLUEMAN AUTOMAGIC SUBNET ####'''


class DhcpdHandler:
    def __init__(self, netconf: "NetConf") -> None:
        self.pid: Optional[int] = None
        self.netconf = netconf

    @staticmethod
    def _read_dhcp_config() -> Tuple[str, str]:
        dhcp_config = ''
        existing_subnet = ''
        start = end = False

        with open(DHCP_CONFIG_FILE) as f:
            for line in f:
                if line == '#### BLUEMAN AUTOMAGIC SUBNET ####\n':
                    start = True
                elif line == '#### END BLUEMAN AUTOMAGIC SUBNET ####\n':
                    if not start:
                        # Because of bug end string got left upon removal
                        continue
                    end = True

                if start and not end:
                    existing_subnet += line
                elif start and end:
                    existing_subnet += line
                    start = end = False
                else:
                    dhcp_config += line

        return dhcp_config, existing_subnet

    def _generate_subnet_config(self) -> str:
        dns = get_dns_servers()

        assert self.netconf.ip4_address is not None and self.netconf.ip4_mask is not None

        ipiface = ipaddress.ip_interface('/'.join((self.netconf.ip4_address, self.netconf.ip4_mask)))

        return DHCPDSUBNET % {"ip_mask": ipiface.network.network_address,
                              "netmask": ipiface.netmask,
                              "dns": dns,
                              "rtr": ipiface.ip,
                              "start": ipiface.network[2],
                              "end": ipiface.network[-2]}

    def do_apply(self) -> None:
        if not self.netconf.locked("dhcp") or self.netconf.ip4_changed:
            if self.netconf.ip4_changed:
                self.do_remove()

            dhcp_config, existing_subnet = self._read_dhcp_config()

            subnet = self._generate_subnet_config()

            # if subnet != self.existing_subnet:
            with open(DHCP_CONFIG_FILE, "w") as f:
                f.write(dhcp_config)
                f.write(subnet)

            cmd = [get_binary("dhcpd3", "dhcpd"), "-pf", "/var/run/dhcpd.pan1.pid", "pan1"]
            p = Popen(cmd, stderr=PIPE)

            error = p.communicate()[1]

            if p.returncode == 0:
                logging.info("dhcpd started correctly")
                with open("/var/run/dhcpd.pan1.pid") as f:
                    self.pid = int(f.read())
                logging.info(f"pid {self.pid}")
                self.netconf.lock("dhcp")
            else:
                error_msg = error.decode("UTF-8").strip()
                logging.info(error_msg)
                raise NetworkSetupError(f"dhcpd failed to start: {error_msg}")

    def do_remove(self) -> None:
        dhcp_config, existing_subnet = self._read_dhcp_config()
        with open(DHCP_CONFIG_FILE, "w") as f:
            f.write(dhcp_config)

        if self.netconf.locked("dhcp"):
            if not self.pid:
                pid = read_pid_file("/var/run/dhcpd.pan1.pid")
            else:
                pid = self.pid

            if not kill(pid, 'dhcpd3') and not kill(pid, 'dhcpd'):
                logging.info("Stale dhcp lockfile found")
            self.netconf.unlock("dhcp")


UDHCP_CONF_TEMPLATE = """start %(start)s
end %(end)s
interface pan1
pidfile /var/run/udhcpd.pan1.pid
option subnet %(ip_mask)s
option dns %(dns)s
option router %(rtr)s
"""


class UdhcpdHandler:
    def __init__(self, netconf: "NetConf") -> None:
        self.pid: Optional[int] = None
        self.netconf = netconf

    def _generate_config(self) -> str:
        dns = get_dns_servers()

        assert self.netconf.ip4_address is not None and self.netconf.ip4_mask is not None

        ipiface = ipaddress.ip_interface('/'.join((self.netconf.ip4_address, self.netconf.ip4_mask)))

        return UDHCP_CONF_TEMPLATE % {"ip_mask": ipiface.network.network_address,
                                      "dns": dns,
                                      "rtr": ipiface.ip,
                                      "start": ipiface.network[2],
                                      "end": ipiface.network[-2]}

    def do_apply(self) -> None:
        if not self.netconf.locked("dhcp") or self.netconf.ip4_changed:
            if self.netconf.ip4_changed:
                self.do_remove()

            config_file, config_path = mkstemp(prefix="udhcpd-")
            os.write(config_file, self._generate_config().encode('UTF-8'))
            os.close(config_file)

            logging.info(f"Running udhcpd with config file {config_path}")
            cmd = [get_binary("udhcpd"), "-S", config_path]
            p = Popen(cmd, stderr=PIPE)
            error = p.communicate()[1]

            # udhcpd takes time to create pid file
            sleep(0.1)

            pid = read_pid_file("/var/run/udhcpd.pan1.pid")

            if p.pid and pid is not None and is_running("udhcpd", pid):
                logging.info("udhcpd started correctly")
                self.pid = pid
                logging.info(f"pid {self.pid}")
                self.netconf.lock("dhcp")
            else:
                error_msg = error.decode("UTF-8").strip()
                logging.info(error_msg)
                raise NetworkSetupError(f"udhcpd failed to start: {error_msg}")

            os.remove(config_path)

    def do_remove(self) -> None:
        if self.netconf.locked("dhcp"):
            if not self.pid:
                pid = read_pid_file("/var/run/udhcpd.pan1.pid")
            else:
                pid = self.pid

            if not kill(pid, 'udhcpd'):
                logging.info("Stale dhcp lockfile found")
            self.netconf.unlock("dhcp")


class_id = 10


class NetConf:
    default_inst = None

    @classmethod
    def get_default(cls) -> "NetConf":
        if NetConf.default_inst:
            return NetConf.default_inst

        try:
            with open("/var/lib/blueman/network.state", "rb") as f:
                obj: "NetConf" = pickle.load(f)
                if obj.version != class_id:
                    raise Exception

                # Convert old 32bit packed ip's to strings
                if not isinstance(obj.ip4_address, str) and obj.ip4_address is not None:
                    obj.ip4_address = str(ipaddress.ip_address(obj.ip4_address))
                    obj.ip4_mask = '255.255.255.0'

                NetConf.default_inst = obj
                return obj
        except (OSError, UnicodeDecodeError, UnpicklingError):
            n = cls()
            try:
                n.store()
            # FIXME find out what could go wrong and drop bare exception
            except:  # noqa: E722
                return n
            NetConf.default_inst = n
            return n

    def __init__(self) -> None:
        self.ip4_mask: Optional[str] = None
        self.ip4_address: Optional[str] = None
        logging.info("init")
        self.version = class_id
        self.dhcp_handler: Optional[DHCPHandler] = None
        self.ipt_rules: List[Tuple[str, str, str]] = []

        self.ip4_changed = False

    def set_ipv4(self, ip: Optional[str], netmask: Optional[str]) -> None:
        if self.ip4_address != ip or self.ip4_mask != netmask:
            self.ip4_changed = True

        self.ip4_address = ip
        self.ip4_mask = netmask

    def get_ipv4(self) -> Tuple[Optional[str], Optional[str]]:
        return self.ip4_address, self.ip4_mask

    @staticmethod
    def enable_ip4_forwarding() -> None:
        with open("/proc/sys/net/ipv4/ip_forward", "w") as f:
            f.write("1")

        for d in os.listdir("/proc/sys/net/ipv4/conf"):
            with open(f"/proc/sys/net/ipv4/conf/{d}/forwarding", "w") as f:
                f.write("1")

    def add_ipt_rule(self, table: str, chain: str, rule: str) -> None:
        self.ipt_rules.append((table, chain, rule))
        args = ["/sbin/iptables", "-t", table, "-A", chain] + rule.split(" ")
        logging.debug(" ".join(args))
        ret = call(args)
        logging.info(f"Return code {ret}")

    def del_ipt_rules(self) -> None:
        for table, chain, rule in self.ipt_rules:
            call(["/sbin/iptables", "-t", table, "-D", chain] + rule.split(" "))
        self.ipt_rules = []
        self.unlock("iptables")

    def set_dhcp_handler(self, handler: Type["DHCPHandler"]) -> None:
        if not isinstance(self.dhcp_handler, handler):
            running = False
            if self.dhcp_handler:
                self.dhcp_handler.do_remove()
                running = True

            self.dhcp_handler = handler(self)
            if running:
                self.dhcp_handler.do_apply()

    def get_dhcp_handler(self) -> Optional[Type["DHCPHandler"]]:
        if not self.dhcp_handler:
            return None

        return type(self.dhcp_handler)

    def apply_settings(self) -> None:
        if self != NetConf.get_default():
            NetConf.get_default().remove_settings()
            NetConf.default_inst = self

        try:
            create_bridge("pan1")
        except BridgeException as e:
            if e.errno != errno.EEXIST:
                raise

        if self.ip4_changed or not self.locked("netconfig"):
            self.enable_ip4_forwarding()

            if have("ip"):
                ret = call(["ip", "link", "set", "dev", "pan1", "up"])
                if ret != 0:
                    raise NetworkSetupError("Failed to bring up interface pan1")

                assert self.ip4_address is not None and self.ip4_mask is not None

                ret = call(["ip", "address", "add", "/".join((self.ip4_address, self.ip4_mask)), "dev", "pan1"])
                if ret != 0:
                    raise NetworkSetupError(f"Failed to add ip address {self.ip4_address}"
                                            f"with netmask {self.ip4_mask}")
            elif have('ifconfig'):
                assert self.ip4_address is not None and self.ip4_mask is not None

                ret = call(["ifconfig", "pan1", self.ip4_address, "netmask", self.ip4_mask, "up"])
                if ret != 0:
                    raise NetworkSetupError(f"Failed to add ip address {self.ip4_address}"
                                            f"with netmask {self.ip4_mask}")
            else:
                raise NetworkSetupError(
                    "Neither ifconfig or ip commands are found. Please install net-tools or iproute2")

            self.lock("netconfig")

        if self.ip4_changed or not self.locked("iptables"):
            self.del_ipt_rules()

            self.add_ipt_rule("nat", "POSTROUTING", f"-s {self.ip4_address}/{self.ip4_mask} -j MASQUERADE")
            self.add_ipt_rule("filter", "FORWARD", "-i pan1 -j ACCEPT")
            self.add_ipt_rule("filter", "FORWARD", "-o pan1 -j ACCEPT")
            self.add_ipt_rule("filter", "FORWARD", "-i pan1 -j ACCEPT")
            self.lock("iptables")

        if self.dhcp_handler:
            self.dhcp_handler.do_apply()
        self.ip4_changed = False

        self.store()

    def remove_settings(self) -> None:
        logging.info(self)

        if self.dhcp_handler:
            self.dhcp_handler.do_remove()

        try:
            destroy_bridge("pan1")
        except BridgeException:
            pass
        self.unlock("netconfig")

        self.del_ipt_rules()

        self.store()

    @staticmethod
    def lock(key: str) -> None:
        with open(f"/var/run/blueman-{key}", "w"):
            pass

    @staticmethod
    def unlock(key: str) -> None:
        try:
            os.unlink(f"/var/run/blueman-{key}")
        except OSError:
            pass

    @staticmethod
    def locked(key: str) -> bool:
        return os.path.exists(f"/var/run/blueman-{key}")

    # save the instance of this class, requires root
    def store(self) -> None:
        if not os.path.exists("/var/lib/blueman"):
            os.mkdir("/var/lib/blueman")
        with open("/var/lib/blueman/network.state", "wb") as f:
            pickle.dump(self, f, 2)



from gettext import gettext as _
import logging

from blueman.main.Builder import Builder
from blueman.plugins.ServicePlugin import ServicePlugin
from blueman.main.DBusProxies import AppletService
from blueman.main.Config import Config

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio


class Transfer(ServicePlugin):
    __plugin_info__ = (_("Transfer"), "document-open")

    def on_load(self, container: Gtk.Box) -> None:

        self._builder = Builder("services-transfer.ui")
        self.widget = self._builder.get_widget("transfer", Gtk.Widget)

        container.pack_start(self.widget, True, True, 0)
        a = AppletService()
        if "TransferService" in a.QueryPlugins():
            self._setup_transfer()
        else:
            self.widget.props.sensitive = False
            self.widget.props.tooltip_text = _("Applet's transfer service plugin is disabled")

    def on_enter(self) -> None:
        self.widget.props.visible = True

    def on_leave(self) -> None:
        self.widget.props.visible = False

    def on_property_changed(self, config: Gio.Settings, key: str) -> None:
        value = config[key]

        if key == "shared-path":
            self._builder.get_widget(key, Gtk.FileChooserButton).set_current_folder(value)
            self.option_changed_notify(key, False)

    def on_apply(self) -> None:
        if self.on_query_apply_state():
            for opt in self.get_options():
                if opt == "shared-path":
                    shared_path = self._builder.get_widget("shared-path", Gtk.FileChooserButton)
                    self._config["shared-path"] = shared_path.get_filename()
                elif opt == "opp-accept":
                    opp_accept = self._builder.get_widget("opp-accept", Gtk.CheckButton)
                    self._config["opp-accept"] = opp_accept.get_active()
                else:
                    raise NotImplementedError("Unknow option: %s" % opt)

            self.clear_options()
            logging.info("transfer apply")

    def on_query_apply_state(self) -> bool:
        opts = self.get_options()
        if not opts:
            return False
        else:
            return True

    def _setup_transfer(self) -> None:
        self._config = Config("org.blueman.transfer")
        self._config.connect("changed", self.on_property_changed)

        opp_accept = self._builder.get_widget("opp-accept", Gtk.CheckButton)
        shared_path = self._builder.get_widget("shared-path", Gtk.FileChooserButton)

        opp_accept.props.active = self._config["opp-accept"]
        if self._config["shared-path"]:
            shared_path.set_current_folder(self._config["shared-path"])

        opp_accept.connect("toggled", lambda x: self.option_changed_notify("opp-accept"))
        shared_path.connect("file-set", lambda x: self.option_changed_notify("shared-path"))



from blueman.bluez.Base import Base
from gi.repository import GLib


class NetworkServer(Base):
    _interface_name = 'org.bluez.NetworkServer1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    def register(self, uuid: str, bridge: str) -> None:
        param = GLib.Variant('(ss)', (uuid, bridge))
        self._call('Register', param)

    def unregister(self, uuid: str) -> None:
        param = GLib.Variant('(s)', (uuid,))
        self._call('Unregister', param)



from typing import List, Callable, Optional, Any, Union, Dict

from gi.repository import Gio, GLib, GObject
from gi.types import GObjectMeta
from blueman.bluez.errors import parse_dbus_error, BluezDBusException
import logging

from blueman.bluemantyping import GSignals


class BaseMeta(GObjectMeta):
    def __call__(cls, *args: object, **kwargs: str) -> "Base":
        if not hasattr(cls, "__instances__"):
            cls.__instances__: Dict[str, "Base"] = {}

        path = kwargs.get('obj_path')
        if path is None:
            path = getattr(cls, "_obj_path")

        if path in cls.__instances__:
            return cls.__instances__[path]

        instance: "Base" = super().__call__(*args, **kwargs)
        cls.__instances__[path] = instance

        return instance


class Base(Gio.DBusProxy, metaclass=BaseMeta):
    connect_signal = GObject.GObject.connect
    disconnect_signal = GObject.GObject.disconnect

    __name = 'org.bluez'
    __bus_type = Gio.BusType.SYSTEM

    __gsignals__: GSignals = {
        'property-changed': (GObject.SignalFlags.NO_HOOKS, None, (str, object, str))
    }
    __instances__: Dict[str, "Base"]

    _interface_name: str

    def __init__(self, *, obj_path: str):
        super().__init__(
            g_name=self.__name,
            g_interface_name=self._interface_name,
            g_object_path=obj_path,
            g_bus_type=self.__bus_type,
            # FIXME See issue 620
            g_flags=Gio.DBusProxyFlags.GET_INVALIDATED_PROPERTIES)

        self.init()
        self.__fallback = {'Icon': 'blueman', 'Class': 0, 'Appearance': 0}

        self.__variant_map = {str: 's', int: 'u', bool: 'b'}

    def do_g_properties_changed(self, changed_properties: GLib.Variant, _invalidated_properties: List[str]) -> None:
        changed = changed_properties.unpack()
        object_path = self.get_object_path()
        logging.debug(f"{object_path} {changed}")
        for key, value in changed.items():
            self.emit("property-changed", key, value, object_path)

    def _call(
        self,
        method: str,
        param: Optional[GLib.Variant] = None,
        reply_handler: Optional[Callable[..., None]] = None,
        error_handler: Optional[Callable[[BluezDBusException], None]] = None,
    ) -> None:
        def callback(
            proxy: Base,
            result: Gio.AsyncResult,
            reply: Optional[Callable[..., None]],
            error: Optional[Callable[[BluezDBusException], None]],
        ) -> None:
            try:
                value = proxy.call_finish(result).unpack()
                if reply:
                    reply(*value)
            except GLib.Error as e:
                if error:
                    error(parse_dbus_error(e))
                else:
                    logging.error(f"Unhandled error for {self.get_interface_name()}.{method}", exc_info=True)

        self.call(method, param, Gio.DBusCallFlags.NONE, GLib.MAXINT, None,
                  callback, reply_handler, error_handler)

    def get(self, name: str) -> Any:
        try:
            prop = self.call_sync(
                'org.freedesktop.DBus.Properties.Get',
                GLib.Variant('(ss)', (self._interface_name, name)),
                Gio.DBusCallFlags.NONE,
                GLib.MAXINT,
                None)
            return prop.unpack()[0]
        except GLib.Error as e:
            property = self.get_cached_property(name)
            if property is not None:
                return property.unpack()
            elif name in self.__fallback:
                return self.__fallback[name]
            else:
                raise parse_dbus_error(e)

    def set(self, name: str, value: Union[str, int, bool]) -> None:
        v = GLib.Variant(self.__variant_map[type(value)], value)
        param = GLib.Variant('(ssv)', (self._interface_name, name, v))
        self.call('org.freedesktop.DBus.Properties.Set',
                  param,
                  Gio.DBusCallFlags.NONE,
                  GLib.MAXINT,
                  None)

    def get_properties(self) -> Dict[str, Any]:
        param = GLib.Variant('(s)', (self._interface_name,))
        res = self.call_sync('org.freedesktop.DBus.Properties.GetAll',
                             param,
                             Gio.DBusCallFlags.NONE,
                             GLib.MAXINT,
                             None)

        props: Dict[str, Any] = res.unpack()[0]
        for k, v in self.__fallback.items():
            if k in props:
                continue
            else:
                props[k] = v

        return props

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __setitem__(self, key: str, value: Union[str, int, bool]) -> None:
        self.set(key, value)

    def __contains__(self, key: str) -> bool:
        return key in self.get_properties()



from datetime import datetime
from gettext import gettext as _
from typing import Optional, overload, TYPE_CHECKING

from blueman.Constants import WEBSITE, VERSION

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

if TYPE_CHECKING:
    from typing_extensions import Literal


class ErrorDialog(Gtk.MessageDialog):
    def __init__(self, markup: str, secondary_markup: Optional[str] = None, excp: Optional[object] = None,
                 icon_name: str = "dialog-error", buttons: Gtk.ButtonsType = Gtk.ButtonsType.CLOSE,
                 title: Optional[str] = None, parent: Optional[Gtk.Container] = None, modal: bool = False,
                 margin_left: int = 0,
                 ) -> None:
        super().__init__(name="ErrorDialog", icon_name=icon_name, buttons=buttons,
                         type=Gtk.MessageType.ERROR, title=title, parent=parent, modal=modal, margin_left=margin_left)

        self.set_markup(markup)

        if secondary_markup:
            self.format_secondary_markup(secondary_markup)

        if excp:
            message_box = self.get_message_area()

            label_expander = Gtk.Label(label="<b>Exception</b>", use_markup=True, visible=True)

            excp_label = Gtk.Label(label=str(excp), selectable=True, visible=True)

            expander = Gtk.Expander(label_widget=label_expander, visible=True)
            expander.add(excp_label)

            message_box.pack_start(expander, False, False, 10)


@overload
def show_about_dialog(app_name: str, run: "Literal[True]" = True, parent: Optional[Gtk.Window] = None) -> None:
    ...


@overload
def show_about_dialog(app_name: str, run: "Literal[False]", parent: Optional[Gtk.Window] = None) -> Gtk.AboutDialog:
    ...


def show_about_dialog(app_name: str, run: bool = True, parent: Optional[Gtk.Window] = None
                      ) -> Optional[Gtk.AboutDialog]:
    about = Gtk.AboutDialog()
    about.set_transient_for(parent)
    about.set_name(app_name)
    about.set_version(VERSION)
    about.set_copyright('Copyright © 2008 Valmantas Palikša\n'
                        'Copyright © 2008 Tadas Dailyda\n'
                        f'Copyright © 2008 - {datetime.now().year} blueman project'
                        )
    about.set_comments(_('Blueman is a GTK+ Bluetooth manager'))
    about.set_website(WEBSITE)
    about.set_website_label(WEBSITE)
    about.set_icon_name('blueman')
    about.set_logo_icon_name('blueman')
    about.set_authors(['Valmantas Palikša <walmis@balticum-tv.lt>',
                       'Tadas Dailyda <tadas@dailyda.com>',
                       f'{WEBSITE}/graphs/contributors'
                       ])
    if run:
        about.run()
        about.destroy()
        return None
    else:
        return about



from typing import TYPE_CHECKING, Optional

from _blueman import device_info
from gi.repository import GLib
from gi.repository import Gtk

from blueman.gui.Animation import Animation
from blueman.gui.manager.ManagerDeviceList import ManagerDeviceList
from blueman.main.SpeedCalc import SpeedCalc
from blueman.Functions import adapter_path_to_name
from blueman.Functions import format_bytes

import gettext

_ = gettext.gettext

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class ManagerStats:
    hbox: Gtk.Box

    def __init__(self, blueman: "Blueman") -> None:

        blueman.List.connect("adapter-changed", self.on_adapter_changed)

        assert blueman.List.Adapter is not None
        self.hci = adapter_path_to_name(blueman.List.Adapter.get_object_path())

        self.time = None

        self.up_speed = SpeedCalc()
        self.down_speed = SpeedCalc()

        self.im_upload = Gtk.Image(icon_name="blueman-up-inactive", pixel_size=16,
                                   halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                   tooltip_text=_("Data activity indication"))
        self.im_download = Gtk.Image(icon_name="blueman-down-inactive", pixel_size=16,
                                     halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                     tooltip_text=_("Data activity indication"))
        self.down_rate = Gtk.Label(halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                   tooltip_text=_("Total data received and rate of transmission"))
        self.down_rate.show()

        self.up_rate = Gtk.Label(halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                 tooltip_text=_("Total data sent and rate of transmission"))
        self.up_rate.show()

        self.uparrow = Gtk.Image(icon_name="go-up", pixel_size=16, halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                 tooltip_text=_("Total data sent and rate of transmission"))
        self.downarrow = Gtk.Image(icon_name="go-down", pixel_size=16, halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
                                   tooltip_text=_("Total data received and rate of transmission"))

        self.hbox = hbox = blueman.builder.get_widget("status_activity", Gtk.Box)

        hbox.pack_start(self.uparrow, False, False, 0)
        hbox.pack_start(self.up_rate, False, False, 0)

        hbox.pack_start(self.downarrow, False, False, 0)
        hbox.pack_start(self.down_rate, False, False, 0)

        hbox.pack_start(Gtk.Separator(orientation=Gtk.Orientation.VERTICAL), False, False, 0)

        hbox.pack_start(self.im_upload, False, False, 0)
        hbox.pack_start(self.im_download, False, False, 0)
        hbox.show_all()
        self.on_adapter_changed(blueman.List, blueman.List.get_adapter_path())

        self.up_blinker = Animation(self.im_upload, ["blueman-up-inactive", "blueman-up-active"])
        self.down_blinker = Animation(self.im_download, ["blueman-down-inactive", "blueman-down-active"])

        self.start_update()

    def on_adapter_changed(self, _lst: ManagerDeviceList, adapter_path: Optional[str]) -> None:
        self.hci = adapter_path_to_name(adapter_path)
        if self.hci is None:
            self.hbox.props.sensitive = False
        else:
            self.hbox.props.sensitive = True

        self.up_speed.reset()
        self.down_speed.reset()

    def set_blinker_by_speed(self, blinker: Animation, speed: float) -> None:

        if speed > 0 and not blinker.status():
            blinker.start()

        if speed > 40 * 1024 and blinker.status():
            blinker.set_rate(10)
        elif speed > (30 * 1024) and blinker.status():
            blinker.set_rate(8)
        elif speed > (20 * 1024) and blinker.status():
            blinker.set_rate(6)
        elif speed > (10 * 1024) and blinker.status():
            blinker.set_rate(4)
        elif speed > 1024 and blinker.status():
            blinker.set_rate(2)
        elif speed == 0 and blinker.status():
            blinker.stop()
        else:
            blinker.set_rate(1)

    def _update(self) -> bool:
        if self.hci is not None:
            devinfo = device_info(self.hci)
            _tx = devinfo["stat"]["byte_tx"]
            _rx = devinfo["stat"]["byte_rx"]

            tx, s_tx = format_bytes(_tx)
            rx, s_rx = format_bytes(_rx)

            _u_speed = self.up_speed.calc(_tx)
            _d_speed = self.down_speed.calc(_rx)

            self.set_blinker_by_speed(self.up_blinker, _u_speed)
            self.set_blinker_by_speed(self.down_blinker, _d_speed)

            u_speed, s_u_speed = format_bytes(_u_speed)
            d_speed, s_d_speed = format_bytes(_d_speed)

            self.set_data(tx, s_tx, rx, s_rx, u_speed, s_u_speed, d_speed, s_d_speed)

        return True

    def start_update(self) -> None:
        self._update()
        self.timer = GLib.timeout_add(1000, self._update)

    def stop_update(self) -> None:
        if self.timer:
            GLib.source_remove(self.timer)

    def set_data(self, uploaded: float, u_name: str, downloaded: float, d_name: str, u_speed: float, us_name: str,
                 d_speed: float, ds_name: str) -> None:
        self.down_rate.set_markup(
            f'<span size="small">{downloaded:.2f} {d_name} <i>{d_speed:5.2f} {ds_name}/s</i></span>')
        self.up_rate.set_markup(
            f'<span size="small">{uploaded:.2f} {u_name} <i>{u_speed:5.2f} {us_name}/s</i></span>')



import builtins
import typing

from gi.repository import GLib


# TODO: Constrain T if possible. Builtins + GTypes might be sufficient?
T = typing.TypeVar('T')

PropertyGetterFn = typing.Callable[[typing.Any], T]
PropertySetterFn = typing.Callable[[typing.Any, T], None]


class Property(typing.Generic[T]):

    name: typing.Optional[str]
    type: typing.Type[T]
    default: typing.Optional[T]
    nick: str
    blurb: str
    flags: ParamFlags
    minimum: typing.Optional[T]
    maximum: typing.Optional[T]

    def __init__(
        self,
        getter: typing.Optional[PropertyGetterFn[T]] = None,
        setter: typing.Optional[PropertySetterFn[T]] = None,
        type: typing.Optional[typing.Type[T]] = None,
        default: typing.Optional[T] = None,
        nick: str = '',
        blurb: str = '',
        flags: ParamFlags = ParamFlags.READWRITE,
        minimum: typing.Optional[T] = None,
        maximum: typing.Optional[T] = None,
    ) -> None:
        ...

    def __get__(self, instance: typing.Any, klass: typing.Type[T]) -> T:
        ...

    def __set__(self, instance: typing.Any, value: T) -> None:
        ...

    def __call__(self, fget: PropertyGetterFn[T]) -> Property[T]:
        ...

    def getter(self: Property[T], fget: PropertyGetterFn[T]) -> Property[T]:
        ...

    def setter(self: Property[T], fset: PropertySetterFn[T]) -> Property[T]:
        ...

    # TODO: There's three Tuple variant structures that could be
    # returned here, and they're all unpleasantly complicated.
    def get_pspec_args(self) -> typing.Sequence[typing.Any]:
        ...


class GType():
    pass

class GInterface():
    ...


class Object():
    Property = Property
    g_type_instance: TypeInstance
    qdata: GLib.Data
    ref_count: builtins.int

    def bind_property(self, source_property: builtins.str, target: Object, target_property: builtins.str, flags: BindingFlags) -> Binding: ...

    def bind_property_full(self, source_property: builtins.str, target: Object, target_property: builtins.str, flags: BindingFlags, transform_to: Closure, transform_from: Closure) -> Binding: ...

    @staticmethod
    def compat_control(what: builtins.int, data: typing.Optional[builtins.object]) -> builtins.int: ...

    T = typing.TypeVar("T")
    T1 = typing.TypeVar("T1")
    T2 = typing.TypeVar("T2")
    T3 = typing.TypeVar("T3")

    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, typing.Any, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, typing.Any, typing.Any, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, typing.Any, typing.Any, typing.Any, typing.Any], typing.Any]) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, T1], typing.Any], user_data1: T1) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, T1], typing.Any], user_data1: T1) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, T1], typing.Any], user_data1: T1) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, typing.Any, typing.Any, T1], typing.Any], user_data1: T1) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, T1, T2], typing.Any], user_data1: T1, user_data2: T2) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, typing.Any, T1, T2], typing.Any], user_data1: T1, user_data2: T2) -> int: ...
    @typing.overload
    def connect(self: T, detailed_signal: str, handler: typing.Callable[[T, T1, T2, T3], typing.Any], user_data1: T1, user_data2: T2, user_data3: T3) -> int: ...

    def connect_after(self, detailed_signal: builtins.str, handler: function, *args: typing.Any) -> builtins.int: ...

    def disconnect(self, handler_id: int) -> None: ...

    def emit(self, signal_name: builtins.str, *args: typing.Any) -> None: ...

    def force_floating(self) -> None: ...

    def freeze_notify(self) -> None: ...

    def get_data(self, key: builtins.str) -> typing.Optional[builtins.object]: ...

    def get_property(self, property_name: builtins.str) -> builtins.object: ...

    def get_qdata(self, quark: builtins.int) -> typing.Optional[builtins.object]: ...

    def getv(self, names: typing.Sequence[builtins.str], values: typing.Sequence[Value]) -> None: ...

    def handler_block(self, handler_id: builtins.int) -> None: ...

    def handler_is_connected(self, handler_id: int) -> bool: ...

    def handler_unblock(self, handler_id: builtins.int) -> None: ...

    @staticmethod
    def interface_find_property(g_iface: TypeInterface, property_name: builtins.str) -> ParamSpec: ...

    @staticmethod
    def interface_install_property(g_iface: TypeInterface, pspec: ParamSpec) -> None: ...

    @staticmethod
    def interface_list_properties(g_iface: TypeInterface) -> typing.Sequence[ParamSpec]: ...

    def is_floating(self) -> builtins.bool: ...

    def notify(self, property_name: builtins.str) -> None: ...

    def notify_by_pspec(self, pspec: ParamSpec) -> None: ...

    def ref(self) -> Object: ...

    def ref_sink(self) -> Object: ...

    def run_dispose(self) -> None: ...

    def set_data(self, key: builtins.str, data: typing.Optional[builtins.object]) -> None: ...

    def set_property(self, property_name: builtins.str, value: builtins.object) -> None: ...

    def set_properties(self, **properties: object) -> None: ...

    def steal_data(self, key: builtins.str) -> typing.Optional[builtins.object]: ...

    def steal_qdata(self, quark: builtins.int) -> typing.Optional[builtins.object]: ...

    def stop_emission(self, detailed_signal: str) -> None: ...

    def thaw_notify(self) -> None: ...

    def unref(self) -> None: ...

    def watch_closure(self, closure: Closure) -> None: ...

    @staticmethod
    def find_property(property_name: builtins.str) -> ParamSpec: ...

    @staticmethod
    def install_properties(pspecs: typing.Sequence[ParamSpec]) -> None: ...

    @staticmethod
    def install_property(property_id: builtins.int, pspec: ParamSpec) -> None: ...

    @staticmethod
    def list_properties() -> typing.Sequence[ParamSpec]: ...

    @staticmethod
    def override_property(property_id: builtins.int, name: builtins.str) -> None: ...

    def do_constructed(self) -> None: ...

    def do_dispatch_properties_changed(self, n_pspecs: builtins.int, pspecs: ParamSpec) -> None: ...

    def do_dispose(self) -> None: ...

    def do_finalize(self) -> None: ...

    def do_get_property(self, property_id: builtins.int, value: Value, pspec: ParamSpec) -> None: ...

    def do_notify(self, pspec: ParamSpec) -> None: ...

    def do_set_property(self, property_id: builtins.int, value: Value, pspec: ParamSpec) -> None: ...


class ParamSpec():
    flags: ParamFlags
    g_type_instance: TypeInstance
    name: builtins.str
    owner_type: GType
    param_id: builtins.int
    qdata: GLib.Data
    ref_count: builtins.int
    value_type: GType

    def get_blurb(self) -> builtins.str: ...

    def get_default_value(self) -> Value: ...

    def get_name(self) -> builtins.str: ...

    def get_name_quark(self) -> builtins.int: ...

    def get_nick(self) -> builtins.str: ...

    def get_qdata(self, quark: builtins.int) -> typing.Optional[builtins.object]: ...

    def get_redirect_target(self) -> ParamSpec: ...

    def set_qdata(self, quark: builtins.int, data: typing.Optional[builtins.object]) -> None: ...

    def sink(self) -> None: ...

    def steal_qdata(self, quark: builtins.int) -> typing.Optional[builtins.object]: ...

    def do_finalize(self) -> None: ...

    def do_value_set_default(self, value: Value) -> None: ...

    def do_value_validate(self, value: Value) -> builtins.bool: ...

    def do_values_cmp(self, value1: Value, value2: Value) -> builtins.int: ...


class TypePlugin(GInterface):

    def complete_interface_info(self, instance_type: GType, interface_type: GType, info: InterfaceInfo) -> None: ...

    def complete_type_info(self, g_type: GType, info: TypeInfo, value_table: TypeValueTable) -> None: ...

    def unuse(self) -> None: ...

    def use(self) -> None: ...


class Binding(Object):

    def get_flags(self) -> BindingFlags: ...

    def get_source(self) -> Object: ...

    def get_source_property(self) -> builtins.str: ...

    def get_target(self) -> Object: ...

    def get_target_property(self) -> builtins.str: ...

    def unbind(self) -> None: ...


class InitiallyUnowned(Object):
    g_type_instance: TypeInstance
    qdata: GLib.Data
    ref_count: builtins.int


class ParamSpecBoolean(ParamSpec):
    default_value: builtins.bool
    parent_instance: ParamSpec


class ParamSpecBoxed(ParamSpec):
    parent_instance: ParamSpec


class ParamSpecChar(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecDouble(ParamSpec):
    default_value: builtins.float
    epsilon: builtins.float
    maximum: builtins.float
    minimum: builtins.float
    parent_instance: ParamSpec


class ParamSpecEnum(ParamSpec):
    default_value: builtins.int
    enum_class: EnumClass
    parent_instance: ParamSpec


class ParamSpecFlags(ParamSpec):
    default_value: builtins.int
    flags_class: FlagsClass
    parent_instance: ParamSpec


class ParamSpecFloat(ParamSpec):
    default_value: builtins.float
    epsilon: builtins.float
    maximum: builtins.float
    minimum: builtins.float
    parent_instance: ParamSpec


class ParamSpecGType(ParamSpec):
    is_a_type: GType
    parent_instance: ParamSpec


class ParamSpecInt(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecInt64(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecLong(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecObject(ParamSpec):
    parent_instance: ParamSpec


class ParamSpecOverride(ParamSpec):
    overridden: ParamSpec
    parent_instance: ParamSpec


class ParamSpecParam(ParamSpec):
    parent_instance: ParamSpec


class ParamSpecPointer(ParamSpec):
    parent_instance: ParamSpec


class ParamSpecString(ParamSpec):
    cset_first: builtins.str
    cset_nth: builtins.str
    default_value: builtins.str
    ensure_non_null: builtins.int
    null_fold_if_empty: builtins.int
    parent_instance: ParamSpec
    substitutor: builtins.int


class ParamSpecUChar(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecUInt(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecUInt64(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecULong(ParamSpec):
    default_value: builtins.int
    maximum: builtins.int
    minimum: builtins.int
    parent_instance: ParamSpec


class ParamSpecUnichar(ParamSpec):
    default_value: builtins.str
    parent_instance: ParamSpec


class ParamSpecValueArray(ParamSpec):
    element_spec: ParamSpec
    fixed_n_elements: builtins.int
    parent_instance: ParamSpec


class ParamSpecVariant(ParamSpec):
    default_value: GLib.Variant
    padding: typing.Sequence[builtins.object]
    parent_instance: ParamSpec
    type: GLib.VariantType


class TypeModule(Object, TypePlugin):
    interface_infos: typing.Sequence[builtins.object]
    name: builtins.str
    parent_instance: Object
    type_infos: typing.Sequence[builtins.object]
    use_count: builtins.int

    def add_interface(self, instance_type: GType, interface_type: GType, interface_info: InterfaceInfo) -> None: ...

    def register_enum(self, name: builtins.str, const_static_values: EnumValue) -> GType: ...

    def register_flags(self, name: builtins.str, const_static_values: FlagsValue) -> GType: ...

    def register_type(self, parent_type: GType, type_name: builtins.str, type_info: TypeInfo, flags: TypeFlags) -> GType: ...

    def set_name(self, name: builtins.str) -> None: ...

    def unuse(self) -> None: ...

    def use(self) -> builtins.bool: ...  # type: ignore

    def do_load(self) -> builtins.bool: ...

    def do_unload(self) -> None: ...


class CClosure():
    callback: builtins.object
    closure: Closure

    @staticmethod
    def marshal_BOOLEAN__BOXED_BOXED(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_BOOLEAN__FLAGS(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_STRING__OBJECT_POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__BOOLEAN(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__BOXED(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__CHAR(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__DOUBLE(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__ENUM(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__FLAGS(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__FLOAT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__INT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__LONG(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__OBJECT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__PARAM(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__STRING(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__UCHAR(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__UINT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__UINT_POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__ULONG(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__VARIANT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_VOID__VOID(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def marshal_generic(closure: Closure, return_gvalue: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


class Closure():
    data: builtins.object
    derivative_flag: builtins.int
    floating: builtins.int
    in_inotify: builtins.int
    in_marshal: builtins.int
    is_invalid: builtins.int
    marshal: builtins.object
    meta_marshal_nouse: builtins.int
    n_fnotifiers: builtins.int
    n_guards: builtins.int
    n_inotifiers: builtins.int
    notifiers: ClosureNotifyData
    ref_count: builtins.int

    def invalidate(self) -> None: ...

    def invoke(self, param_values: typing.Sequence[Value], invocation_hint: typing.Optional[builtins.object]) -> Value: ...

    @staticmethod
    def new_object(sizeof_closure: builtins.int, object: Object) -> Closure: ...

    @staticmethod
    def new_simple(sizeof_closure: builtins.int, data: typing.Optional[builtins.object]) -> Closure: ...

    def ref(self) -> Closure: ...

    def sink(self) -> None: ...

    def unref(self) -> None: ...


class ClosureNotifyData():
    data: builtins.object
    notify: ClosureNotify


class EnumClass():
    g_type_class: TypeClass
    maximum: builtins.int
    minimum: builtins.int
    n_values: builtins.int
    values: EnumValue


class EnumValue():
    value: builtins.int
    value_name: builtins.str
    value_nick: builtins.str


class FlagsClass():
    g_type_class: TypeClass
    mask: builtins.int
    n_values: builtins.int
    values: FlagsValue


class FlagsValue():
    value: builtins.int
    value_name: builtins.str
    value_nick: builtins.str


class InterfaceInfo():
    interface_data: builtins.object
    interface_finalize: InterfaceFinalizeFunc
    interface_init: InterfaceInitFunc


class ObjectConstructParam():
    pspec: ParamSpec
    value: Value


class ParamSpecPool():

    def insert(self, pspec: ParamSpec, owner_type: GType) -> None: ...

    def list(self, owner_type: GType) -> typing.Sequence[ParamSpec]: ...

    def list_owned(self, owner_type: GType) -> typing.Sequence[ParamSpec]: ...

    def lookup(self, param_name: builtins.str, owner_type: GType, walk_ancestors: builtins.bool) -> ParamSpec: ...

    @staticmethod
    def new(type_prefixing: builtins.bool) -> ParamSpecPool: ...

    def remove(self, pspec: ParamSpec) -> None: ...


class ParamSpecTypeInfo():
    finalize: builtins.object
    instance_init: builtins.object
    instance_size: builtins.int
    n_preallocs: builtins.int
    value_set_default: builtins.object
    value_type: GType
    value_validate: builtins.object
    values_cmp: builtins.object


class Parameter():
    name: builtins.str
    value: Value


class SignalInvocationHint():
    detail: builtins.int
    run_type: SignalFlags
    signal_id: builtins.int


class SignalQuery():
    itype: GType
    n_params: builtins.int
    param_types: typing.Sequence[GType]
    return_type: GType
    signal_flags: SignalFlags
    signal_id: builtins.int
    signal_name: builtins.str


class TypeClass():
    g_type: GType

    def add_private(self, private_size: builtins.int) -> None: ...

    @staticmethod
    def adjust_private_offset(g_class: typing.Optional[builtins.object], private_size_or_offset: builtins.int) -> None: ...

    def get_private(self, private_type: GType) -> typing.Optional[builtins.object]: ...

    @staticmethod
    def peek(type: GType) -> TypeClass: ...

    def peek_parent(self) -> TypeClass: ...

    @staticmethod
    def peek_static(type: GType) -> TypeClass: ...

    @staticmethod
    def ref(type: GType) -> TypeClass: ...

    def unref(self) -> None: ...


class TypeFundamentalInfo():
    type_flags: TypeFundamentalFlags


class TypeInfo():
    base_finalize: BaseFinalizeFunc
    base_init: BaseInitFunc
    class_data: builtins.object
    class_finalize: ClassFinalizeFunc
    class_init: ClassInitFunc
    class_size: builtins.int
    instance_init: InstanceInitFunc
    instance_size: builtins.int
    n_preallocs: builtins.int
    value_table: TypeValueTable


class TypeInstance():
    g_class: TypeClass

    def get_private(self, private_type: GType) -> typing.Optional[builtins.object]: ...


class TypeInterface():
    g_instance_type: GType
    g_type: GType

    @staticmethod
    def add_prerequisite(interface_type: GType, prerequisite_type: GType) -> None: ...

    @staticmethod
    def get_plugin(instance_type: GType, interface_type: GType) -> TypePlugin: ...

    @staticmethod
    def peek(instance_class: TypeClass, iface_type: GType) -> TypeInterface: ...

    def peek_parent(self) -> TypeInterface: ...

    @staticmethod
    def prerequisites(interface_type: GType) -> typing.Sequence[GType]: ...


class TypePluginClass():
    base_iface: TypeInterface
    complete_interface_info: TypePluginCompleteInterfaceInfo
    complete_type_info: TypePluginCompleteTypeInfo
    unuse_plugin: TypePluginUnuse
    use_plugin: TypePluginUse


class TypeQuery():
    class_size: builtins.int
    instance_size: builtins.int
    type: GType
    type_name: builtins.str


class TypeValueTable():
    collect_format: builtins.str
    collect_value: builtins.object
    lcopy_format: builtins.str
    lcopy_value: builtins.object
    value_copy: builtins.object
    value_free: builtins.object
    value_init: builtins.object
    value_peek_pointer: builtins.object


class Value():
    g_type: GType

    def copy(self, dest_value: Value) -> None: ...

    def dup_object(self) -> Object: ...

    def dup_string(self) -> builtins.str: ...

    def dup_variant(self) -> typing.Optional[GLib.Variant]: ...

    def fits_pointer(self) -> builtins.bool: ...

    def get_boolean(self) -> builtins.bool: ...

    def get_boxed(self) -> typing.Optional[builtins.object]: ...

    def get_char(self) -> builtins.int: ...

    def get_double(self) -> builtins.float: ...

    def get_enum(self) -> builtins.int: ...

    def get_flags(self) -> builtins.int: ...

    def get_float(self) -> builtins.float: ...

    def get_gtype(self) -> GType: ...

    def get_int(self) -> builtins.int: ...

    def get_int64(self) -> builtins.int: ...

    def get_long(self) -> builtins.int: ...

    def get_object(self) -> Object: ...

    def get_param(self) -> ParamSpec: ...

    def get_pointer(self) -> typing.Optional[builtins.object]: ...

    def get_schar(self) -> builtins.int: ...

    def get_string(self) -> builtins.str: ...

    def get_uchar(self) -> builtins.int: ...

    def get_uint(self) -> builtins.int: ...

    def get_uint64(self) -> builtins.int: ...

    def get_ulong(self) -> builtins.int: ...

    def get_variant(self) -> typing.Optional[GLib.Variant]: ...

    def init(self, g_type: GType) -> Value: ...

    def init_from_instance(self, instance: TypeInstance) -> None: ...

    def peek_pointer(self) -> typing.Optional[builtins.object]: ...

    def reset(self) -> Value: ...

    def set_boolean(self, v_boolean: builtins.bool) -> None: ...

    def set_boxed(self, v_boxed: typing.Optional[builtins.object]) -> None: ...

    def set_boxed_take_ownership(self, v_boxed: typing.Optional[builtins.object]) -> None: ...

    def set_char(self, v_char: builtins.int) -> None: ...

    def set_double(self, v_double: builtins.float) -> None: ...

    def set_enum(self, v_enum: builtins.int) -> None: ...

    def set_flags(self, v_flags: builtins.int) -> None: ...

    def set_float(self, v_float: builtins.float) -> None: ...

    def set_gtype(self, v_gtype: GType) -> None: ...

    def set_instance(self, instance: typing.Optional[builtins.object]) -> None: ...

    def set_int(self, v_int: builtins.int) -> None: ...

    def set_int64(self, v_int64: builtins.int) -> None: ...

    def set_long(self, v_long: builtins.int) -> None: ...

    def set_object(self, v_object: typing.Optional[Object]) -> None: ...

    def set_param(self, param: typing.Optional[ParamSpec]) -> None: ...

    def set_pointer(self, v_pointer: typing.Optional[builtins.object]) -> None: ...

    def set_schar(self, v_char: builtins.int) -> None: ...

    def set_static_boxed(self, v_boxed: typing.Optional[builtins.object]) -> None: ...

    def set_static_string(self, v_string: typing.Optional[builtins.str]) -> None: ...

    def set_string(self, v_string: typing.Optional[builtins.str]) -> None: ...

    def set_string_take_ownership(self, v_string: typing.Optional[builtins.str]) -> None: ...

    def set_uchar(self, v_uchar: builtins.int) -> None: ...

    def set_uint(self, v_uint: builtins.int) -> None: ...

    def set_uint64(self, v_uint64: builtins.int) -> None: ...

    def set_ulong(self, v_ulong: builtins.int) -> None: ...

    def set_variant(self, variant: typing.Optional[GLib.Variant]) -> None: ...

    def take_boxed(self, v_boxed: typing.Optional[builtins.object]) -> None: ...

    def take_string(self, v_string: typing.Optional[builtins.str]) -> None: ...

    def take_variant(self, variant: typing.Optional[GLib.Variant]) -> None: ...

    def transform(self, dest_value: Value) -> builtins.bool: ...

    @staticmethod
    def type_compatible(src_type: GType, dest_type: GType) -> builtins.bool: ...

    @staticmethod
    def type_transformable(src_type: GType, dest_type: GType) -> builtins.bool: ...

    def unset(self) -> None: ...


class ValueArray():
    n_prealloced: builtins.int
    n_values: builtins.int
    values: Value

    def append(self, value: typing.Optional[Value]) -> ValueArray: ...

    def copy(self) -> ValueArray: ...

    def get_nth(self, index_: builtins.int) -> Value: ...

    def insert(self, index_: builtins.int, value: typing.Optional[Value]) -> ValueArray: ...

    @staticmethod
    def new(n_prealloced: builtins.int) -> ValueArray: ...

    def prepend(self, value: typing.Optional[Value]) -> ValueArray: ...

    def remove(self, index_: builtins.int) -> ValueArray: ...

    def sort(self, compare_func: GLib.CompareDataFunc, *user_data: typing.Optional[builtins.object]) -> ValueArray: ...


class WeakRef():
    ...


class TypeCValue():
    ...


class BindingFlags(GFlags, builtins.int):
    BIDIRECTIONAL = ...  # type: BindingFlags
    DEFAULT = ...  # type: BindingFlags
    INVERT_BOOLEAN = ...  # type: BindingFlags
    SYNC_CREATE = ...  # type: BindingFlags


class ConnectFlags(GLib.Flags, builtins.int):
    AFTER = ...  # type: ConnectFlags
    SWAPPED = ...  # type: ConnectFlags


class GFlags(GLib.Flags, builtins.int):
    ...


class ParamFlags(GLib.Flags, builtins.int):
    CONSTRUCT = ...  # type: ParamFlags
    CONSTRUCT_ONLY = ...  # type: ParamFlags
    DEPRECATED = ...  # type: ParamFlags
    EXPLICIT_NOTIFY = ...  # type: ParamFlags
    LAX_VALIDATION = ...  # type: ParamFlags
    PRIVATE = ...  # type: ParamFlags
    READABLE = ...  # type: ParamFlags
    READWRITE = ...  # type: ParamFlags
    STATIC_BLURB = ...  # type: ParamFlags
    STATIC_NAME = ...  # type: ParamFlags
    STATIC_NICK = ...  # type: ParamFlags
    WRITABLE = ...  # type: ParamFlags


class SignalFlags(GLib.Flags, builtins.int):
    ACTION = ...  # type: SignalFlags
    DEPRECATED = ...  # type: SignalFlags
    DETAILED = ...  # type: SignalFlags
    MUST_COLLECT = ...  # type: SignalFlags
    NO_HOOKS = ...  # type: SignalFlags
    NO_RECURSE = ...  # type: SignalFlags
    RUN_CLEANUP = ...  # type: SignalFlags
    RUN_FIRST = ...  # type: SignalFlags
    RUN_LAST = ...  # type: SignalFlags


class SignalMatchType(GLib.Flags, builtins.int):
    CLOSURE = ...  # type: SignalMatchType
    DATA = ...  # type: SignalMatchType
    DETAIL = ...  # type: SignalMatchType
    FUNC = ...  # type: SignalMatchType
    ID = ...  # type: SignalMatchType
    UNBLOCKED = ...  # type: SignalMatchType


class TypeDebugFlags(GLib.Flags, builtins.int):
    INSTANCE_COUNT = ...  # type: TypeDebugFlags
    MASK = ...  # type: TypeDebugFlags
    NONE = ...  # type: TypeDebugFlags
    OBJECTS = ...  # type: TypeDebugFlags
    SIGNALS = ...  # type: TypeDebugFlags


class TypeFlags(GLib.Flags, builtins.int):
    ABSTRACT = ...  # type: TypeFlags
    VALUE_ABSTRACT = ...  # type: TypeFlags


class TypeFundamentalFlags(GLib.Flags, builtins.int):
    CLASSED = ...  # type: TypeFundamentalFlags
    DEEP_DERIVABLE = ...  # type: TypeFundamentalFlags
    DERIVABLE = ...  # type: TypeFundamentalFlags
    INSTANTIATABLE = ...  # type: TypeFundamentalFlags


class GEnum(GLib.Enum, builtins.int):
    ...


BaseFinalizeFunc = typing.Callable[[TypeClass], None]
BaseInitFunc = typing.Callable[[TypeClass], None]
BindingTransformFunc = typing.Callable[[Binding, Value, Value, typing.Optional[builtins.object]], builtins.bool]
BoxedCopyFunc = typing.Callable[[builtins.object], builtins.object]
BoxedFreeFunc = typing.Callable[[builtins.object], None]
Callback = typing.Callable[[], None]
ClassFinalizeFunc = typing.Callable[[TypeClass, typing.Optional[builtins.object]], None]
ClassInitFunc = typing.Callable[[TypeClass, typing.Optional[builtins.object]], None]
ClosureMarshal = typing.Callable[[Closure, typing.Optional[Value], typing.Sequence[Value], typing.Optional[builtins.object], typing.Optional[builtins.object]], None]
ClosureNotify = typing.Callable[[typing.Optional[builtins.object], Closure], None]
InstanceInitFunc = typing.Callable[[TypeInstance, TypeClass], None]
InterfaceFinalizeFunc = typing.Callable[[TypeInterface, typing.Optional[builtins.object]], None]
InterfaceInitFunc = typing.Callable[[TypeInterface, typing.Optional[builtins.object]], None]
ObjectFinalizeFunc = typing.Callable[[Object], None]
ObjectGetPropertyFunc = typing.Callable[[Object, builtins.int, Value, ParamSpec], None]
ObjectSetPropertyFunc = typing.Callable[[Object, builtins.int, Value, ParamSpec], None]
SignalAccumulator = typing.Callable[[SignalInvocationHint, Value, Value, typing.Optional[builtins.object]], builtins.bool]
SignalEmissionHook = typing.Callable[[SignalInvocationHint, typing.Sequence[Value], typing.Optional[builtins.object]], builtins.bool]
ToggleNotify = typing.Callable[[typing.Optional[builtins.object], Object, builtins.bool], None]
TypeClassCacheFunc = typing.Callable[[typing.Optional[builtins.object], TypeClass], builtins.bool]
TypeInterfaceCheckFunc = typing.Callable[[typing.Optional[builtins.object], TypeInterface], None]
TypePluginCompleteInterfaceInfo = typing.Callable[[TypePlugin, GType, GType, InterfaceInfo], None]
TypePluginCompleteTypeInfo = typing.Callable[[TypePlugin, GType, TypeInfo, TypeValueTable], None]
TypePluginUnuse = typing.Callable[[TypePlugin], None]
TypePluginUse = typing.Callable[[TypePlugin], None]
ValueTransform = typing.Callable[[Value, Value], None]
WeakNotify = typing.Callable[[typing.Optional[builtins.object], Object], None]


def boxed_copy(boxed_type: GType, src_boxed: builtins.object) -> builtins.object: ...


def boxed_free(boxed_type: GType, boxed: builtins.object) -> None: ...


def cclosure_marshal_BOOLEAN__BOXED_BOXED(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_BOOLEAN__FLAGS(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_STRING__OBJECT_POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__BOOLEAN(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__BOXED(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__CHAR(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__DOUBLE(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__ENUM(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__FLAGS(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__FLOAT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__INT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__LONG(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__OBJECT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__PARAM(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__STRING(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__UCHAR(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__UINT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__UINT_POINTER(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__ULONG(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__VARIANT(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_VOID__VOID(closure: Closure, return_value: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def cclosure_marshal_generic(closure: Closure, return_gvalue: Value, n_param_values: builtins.int, param_values: Value, invocation_hint: typing.Optional[builtins.object], marshal_data: typing.Optional[builtins.object]) -> None: ...


def clear_signal_handler(handler_id_ptr: builtins.int, instance: Object) -> None: ...


def enum_complete_type_info(g_enum_type: GType, const_values: EnumValue) -> TypeInfo: ...


def enum_get_value(enum_class: EnumClass, value: builtins.int) -> EnumValue: ...


def enum_get_value_by_name(enum_class: EnumClass, name: builtins.str) -> EnumValue: ...


def enum_get_value_by_nick(enum_class: EnumClass, nick: builtins.str) -> EnumValue: ...


def enum_register_static(name: builtins.str, const_static_values: EnumValue) -> GType: ...


def enum_to_string(g_enum_type: GType, value: builtins.int) -> builtins.str: ...


def flags_complete_type_info(g_flags_type: GType, const_values: FlagsValue) -> TypeInfo: ...


def flags_get_first_value(flags_class: FlagsClass, value: builtins.int) -> FlagsValue: ...


def flags_get_value_by_name(flags_class: FlagsClass, name: builtins.str) -> FlagsValue: ...


def flags_get_value_by_nick(flags_class: FlagsClass, nick: builtins.str) -> FlagsValue: ...


def flags_register_static(name: builtins.str, const_static_values: FlagsValue) -> GType: ...


def flags_to_string(flags_type: GType, value: builtins.int) -> builtins.str: ...


def gtype_get_type() -> GType: ...


def param_spec_boolean(name: builtins.str, nick: builtins.str, blurb: builtins.str, default_value: builtins.bool, flags: ParamFlags) -> ParamSpec: ...


def param_spec_boxed(name: builtins.str, nick: builtins.str, blurb: builtins.str, boxed_type: GType, flags: ParamFlags) -> ParamSpec: ...


def param_spec_char(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_double(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.float, maximum: builtins.float, default_value: builtins.float, flags: ParamFlags) -> ParamSpec: ...


def param_spec_enum(name: builtins.str, nick: builtins.str, blurb: builtins.str, enum_type: GType, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_flags(name: builtins.str, nick: builtins.str, blurb: builtins.str, flags_type: GType, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_float(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.float, maximum: builtins.float, default_value: builtins.float, flags: ParamFlags) -> ParamSpec: ...


def param_spec_gtype(name: builtins.str, nick: builtins.str, blurb: builtins.str, is_a_type: GType, flags: ParamFlags) -> ParamSpec: ...


def param_spec_int(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_int64(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_long(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_object(name: builtins.str, nick: builtins.str, blurb: builtins.str, object_type: GType, flags: ParamFlags) -> ParamSpec: ...


def param_spec_param(name: builtins.str, nick: builtins.str, blurb: builtins.str, param_type: GType, flags: ParamFlags) -> ParamSpec: ...


def param_spec_pointer(name: builtins.str, nick: builtins.str, blurb: builtins.str, flags: ParamFlags) -> ParamSpec: ...


def param_spec_pool_new(type_prefixing: builtins.bool) -> ParamSpecPool: ...


def param_spec_string(name: builtins.str, nick: builtins.str, blurb: builtins.str, default_value: typing.Optional[builtins.str], flags: ParamFlags) -> ParamSpec: ...


def param_spec_uchar(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_uint(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_uint64(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_ulong(name: builtins.str, nick: builtins.str, blurb: builtins.str, minimum: builtins.int, maximum: builtins.int, default_value: builtins.int, flags: ParamFlags) -> ParamSpec: ...


def param_spec_unichar(name: builtins.str, nick: builtins.str, blurb: builtins.str, default_value: builtins.str, flags: ParamFlags) -> ParamSpec: ...


def param_spec_variant(name: builtins.str, nick: builtins.str, blurb: builtins.str, type: GLib.VariantType, default_value: typing.Optional[GLib.Variant], flags: ParamFlags) -> ParamSpec: ...


def param_type_register_static(name: builtins.str, pspec_info: ParamSpecTypeInfo) -> GType: ...


def param_value_convert(pspec: ParamSpec, src_value: Value, dest_value: Value, strict_validation: builtins.bool) -> builtins.bool: ...


def param_value_defaults(pspec: ParamSpec, value: Value) -> builtins.bool: ...


def param_value_set_default(pspec: ParamSpec, value: Value) -> None: ...


def param_value_validate(pspec: ParamSpec, value: Value) -> builtins.bool: ...


def param_values_cmp(pspec: ParamSpec, value1: Value, value2: Value) -> builtins.int: ...


def pointer_type_register_static(name: builtins.str) -> GType: ...


def signal_accumulator_first_wins(ihint: SignalInvocationHint, return_accu: Value, handler_return: Value, dummy: typing.Optional[builtins.object]) -> builtins.bool: ...


def signal_accumulator_true_handled(ihint: SignalInvocationHint, return_accu: Value, handler_return: Value, dummy: typing.Optional[builtins.object]) -> builtins.bool: ...


def signal_add_emission_hook(signal_id: builtins.int, detail: builtins.int, hook_func: SignalEmissionHook, *hook_data: typing.Optional[builtins.object]) -> builtins.int: ...


def signal_chain_from_overridden(instance_and_params: typing.Sequence[Value], return_value: Value) -> None: ...


def signal_connect_closure(instance: Object, detailed_signal: builtins.str, closure: Closure, after: builtins.bool) -> builtins.int: ...


def signal_connect_closure_by_id(instance: Object, signal_id: builtins.int, detail: builtins.int, closure: Closure, after: builtins.bool) -> builtins.int: ...


def signal_emitv(instance_and_params: typing.Sequence[Value], signal_id: builtins.int, detail: builtins.int, return_value: Value) -> Value: ...


def signal_get_invocation_hint(instance: Object) -> SignalInvocationHint: ...


def signal_handler_block(instance: Object, handler_id: builtins.int) -> None: ...


def signal_handler_disconnect(instance: Object, handler_id: builtins.int) -> None: ...


def signal_handler_find(instance: Object, mask: SignalMatchType, signal_id: builtins.int, detail: builtins.int, closure: typing.Optional[Closure], func: typing.Optional[builtins.object], data: typing.Optional[builtins.object]) -> builtins.int: ...


def signal_handler_is_connected(instance: Object, handler_id: builtins.int) -> builtins.bool: ...


def signal_handler_unblock(instance: Object, handler_id: builtins.int) -> None: ...


def signal_handlers_block_matched(instance: Object, mask: SignalMatchType, signal_id: builtins.int, detail: builtins.int, closure: typing.Optional[Closure], func: typing.Optional[builtins.object], data: typing.Optional[builtins.object]) -> builtins.int: ...


def signal_handlers_destroy(instance: Object) -> None: ...


def signal_handlers_disconnect_matched(instance: Object, mask: SignalMatchType, signal_id: builtins.int, detail: builtins.int, closure: typing.Optional[Closure], func: typing.Optional[builtins.object], data: typing.Optional[builtins.object]) -> builtins.int: ...


def signal_handlers_unblock_matched(instance: Object, mask: SignalMatchType, signal_id: builtins.int, detail: builtins.int, closure: typing.Optional[Closure], func: typing.Optional[builtins.object], data: typing.Optional[builtins.object]) -> builtins.int: ...


def signal_has_handler_pending(instance: Object, signal_id: builtins.int, detail: builtins.int, may_be_blocked: builtins.bool) -> builtins.bool: ...


def signal_list_ids(itype: GType) -> typing.Sequence[builtins.int]: ...


def signal_lookup(name: builtins.str, itype: GType) -> builtins.int: ...


def signal_name(signal_id: builtins.int) -> builtins.str: ...


def signal_override_class_closure(signal_id: builtins.int, instance_type: GType, class_closure: Closure) -> None: ...


def signal_parse_name(detailed_signal: builtins.str, itype: GType, force_detail_quark: builtins.bool) -> typing.Tuple[builtins.bool, builtins.int, builtins.int]: ...


def signal_query(signal_id: builtins.int) -> SignalQuery: ...


def signal_remove_emission_hook(signal_id: builtins.int, hook_id: builtins.int) -> None: ...


def signal_stop_emission(instance: Object, signal_id: builtins.int, detail: builtins.int) -> None: ...


def signal_stop_emission_by_name(instance: Object, detailed_signal: builtins.str) -> None: ...


def signal_type_cclosure_new(itype: GType, struct_offset: builtins.int) -> Closure: ...


def source_set_closure(source: GLib.Source, closure: Closure) -> None: ...


def source_set_dummy_callback(source: GLib.Source) -> None: ...


def strdup_value_contents(value: Value) -> builtins.str: ...


def type_add_class_private(class_type: GType, private_size: builtins.int) -> None: ...


def type_add_instance_private(class_type: GType, private_size: builtins.int) -> builtins.int: ...


def type_add_interface_dynamic(instance_type: GType, interface_type: GType, plugin: TypePlugin) -> None: ...


def type_add_interface_static(instance_type: GType, interface_type: GType, info: InterfaceInfo) -> None: ...


def type_check_class_is_a(g_class: TypeClass, is_a_type: GType) -> builtins.bool: ...


def type_check_instance(instance: TypeInstance) -> builtins.bool: ...


def type_check_instance_is_a(instance: TypeInstance, iface_type: GType) -> builtins.bool: ...


def type_check_instance_is_fundamentally_a(instance: TypeInstance, fundamental_type: GType) -> builtins.bool: ...


def type_check_is_value_type(type: GType) -> builtins.bool: ...


def type_check_value(value: Value) -> builtins.bool: ...


def type_check_value_holds(value: Value, type: GType) -> builtins.bool: ...


def type_children(type: GType) -> typing.Sequence[GType]: ...


def type_class_adjust_private_offset(g_class: typing.Optional[builtins.object], private_size_or_offset: builtins.int) -> None: ...


def type_class_peek(type: GType) -> TypeClass: ...


def type_class_peek_static(type: GType) -> TypeClass: ...


def type_class_ref(type: GType) -> TypeClass: ...


def type_default_interface_peek(g_type: GType) -> TypeInterface: ...


def type_default_interface_ref(g_type: GType) -> TypeInterface: ...


def type_default_interface_unref(g_iface: TypeInterface) -> None: ...


def type_depth(type: GType) -> builtins.int: ...


def type_ensure(type: GType) -> None: ...


def type_free_instance(instance: TypeInstance) -> None: ...


def type_from_name(name: builtins.str) -> GType: ...


def type_fundamental(type_id: GType) -> GType: ...


def type_fundamental_next() -> GType: ...


def type_get_instance_count(type: GType) -> builtins.int: ...


def type_get_plugin(type: GType) -> TypePlugin: ...


def type_get_qdata(type: GType, quark: builtins.int) -> typing.Optional[builtins.object]: ...


def type_get_type_registration_serial() -> builtins.int: ...


def type_init() -> None: ...


def type_init_with_debug_flags(debug_flags: TypeDebugFlags) -> None: ...


def type_interface_add_prerequisite(interface_type: GType, prerequisite_type: GType) -> None: ...


def type_interface_get_plugin(instance_type: GType, interface_type: GType) -> TypePlugin: ...


def type_interface_peek(instance_class: TypeClass, iface_type: GType) -> TypeInterface: ...


def type_interface_prerequisites(interface_type: GType) -> typing.Sequence[GType]: ...


def type_interfaces(type: GType) -> typing.Sequence[GType]: ...


def type_is_a(type: GType, is_a_type: GType) -> builtins.bool: ...


def type_name(type: GType) -> builtins.str: ...


def type_name_from_class(g_class: TypeClass) -> builtins.str: ...


def type_name_from_instance(instance: TypeInstance) -> builtins.str: ...


def type_next_base(leaf_type: GType, root_type: GType) -> GType: ...


def type_parent(type: GType) -> GType: ...


def type_qname(type: GType) -> builtins.int: ...


def type_query(type: GType) -> TypeQuery: ...


def type_register_dynamic(parent_type: GType, type_name: builtins.str, plugin: TypePlugin, flags: TypeFlags) -> GType: ...


def type_register_fundamental(type_id: GType, type_name: builtins.str, info: TypeInfo, finfo: TypeFundamentalInfo, flags: TypeFlags) -> GType: ...


def type_register_static(parent_type: GType, type_name: builtins.str, info: TypeInfo, flags: TypeFlags) -> GType: ...


def type_set_qdata(type: GType, quark: builtins.int, data: typing.Optional[builtins.object]) -> None: ...


def type_test_flags(type: GType, flags: builtins.int) -> builtins.bool: ...


def value_type_compatible(src_type: GType, dest_type: GType) -> builtins.bool: ...


def value_type_transformable(src_type: GType, dest_type: GType) -> builtins.bool: ...


GBoxed: typing.Any
GObjectWeakRef: typing.Any
GParamSpec: typing.Any
GPointer: typing.Any
G_MAXDOUBLE: builtins.float
G_MAXFLOAT: builtins.float
G_MAXINT: builtins.int
G_MAXINT16: builtins.int
G_MAXINT32: builtins.int
G_MAXINT64: builtins.int
G_MAXINT8: builtins.int
G_MAXLONG: builtins.int
G_MAXOFFSET: builtins.int
G_MAXSHORT: builtins.int
G_MAXSIZE: builtins.int
G_MAXSSIZE: builtins.int
G_MAXUINT: builtins.int
G_MAXUINT16: builtins.int
G_MAXUINT32: builtins.int
G_MAXUINT64: builtins.int
G_MAXUINT8: builtins.int
G_MAXULONG: builtins.int
G_MAXUSHORT: builtins.int
G_MINDOUBLE: builtins.float
G_MINFLOAT: builtins.float
G_MININT: builtins.int
G_MININT16: builtins.int
G_MININT32: builtins.int
G_MININT64: builtins.int
G_MININT8: builtins.int
G_MINLONG: builtins.int
G_MINOFFSET: builtins.int
G_MINSHORT: builtins.int
G_MINSSIZE: builtins.int
IO_ERR: GLib.IOCondition
IO_FLAG_APPEND: GLib.IOFlags
IO_FLAG_GET_MASK: GLib.IOFlags
IO_FLAG_IS_READABLE: GLib.IOFlags
IO_FLAG_IS_SEEKABLE: GLib.IOFlags
IO_FLAG_IS_WRITEABLE: GLib.IOFlags
IO_FLAG_MASK: GLib.IOFlags
IO_FLAG_NONBLOCK: GLib.IOFlags
IO_FLAG_SET_MASK: GLib.IOFlags
IO_HUP: GLib.IOCondition
IO_IN: GLib.IOCondition
IO_NVAL: GLib.IOCondition
IO_OUT: GLib.IOCondition
IO_PRI: GLib.IOCondition
IO_STATUS_AGAIN: GLib.IOStatus
IO_STATUS_EOF: GLib.IOStatus
IO_STATUS_ERROR: GLib.IOStatus
IO_STATUS_NORMAL: GLib.IOStatus
OPTION_ERROR_BAD_VALUE: GLib.OptionError
OPTION_ERROR_FAILED: GLib.OptionError
OPTION_ERROR_UNKNOWN_OPTION: GLib.OptionError
OPTION_FLAG_FILENAME: GLib.OptionFlags
OPTION_FLAG_HIDDEN: GLib.OptionFlags
OPTION_FLAG_IN_MAIN: GLib.OptionFlags
OPTION_FLAG_NOALIAS: GLib.OptionFlags
OPTION_FLAG_NO_ARG: GLib.OptionFlags
OPTION_FLAG_OPTIONAL_ARG: GLib.OptionFlags
OPTION_FLAG_REVERSE: GLib.OptionFlags
OPTION_REMAINING: builtins.str
PARAM_CONSTRUCT: ParamFlags
PARAM_CONSTRUCT_ONLY: ParamFlags
PARAM_LAX_VALIDATION: ParamFlags
PARAM_MASK: builtins.int
PARAM_READABLE: ParamFlags
PARAM_READWRITE: ParamFlags
PARAM_STATIC_STRINGS: builtins.int
PARAM_USER_SHIFT: builtins.int
PARAM_WRITABLE: ParamFlags
PRIORITY_DEFAULT: builtins.int
PRIORITY_DEFAULT_IDLE: builtins.int
PRIORITY_HIGH: builtins.int
PRIORITY_HIGH_IDLE: builtins.int
PRIORITY_LOW: builtins.int
SIGNAL_ACTION: SignalFlags
SIGNAL_DETAILED: SignalFlags
SIGNAL_FLAGS_MASK: builtins.int
SIGNAL_MATCH_MASK: builtins.int
SIGNAL_NO_HOOKS: SignalFlags
SIGNAL_NO_RECURSE: SignalFlags
SIGNAL_RUN_CLEANUP: SignalFlags
SIGNAL_RUN_FIRST: SignalFlags
SIGNAL_RUN_LAST: SignalFlags
SPAWN_CHILD_INHERITS_STDIN: GLib.SpawnFlags
SPAWN_DO_NOT_REAP_CHILD: GLib.SpawnFlags
SPAWN_FILE_AND_ARGV_ZERO: GLib.SpawnFlags
SPAWN_LEAVE_DESCRIPTORS_OPEN: GLib.SpawnFlags
SPAWN_SEARCH_PATH: GLib.SpawnFlags
SPAWN_STDERR_TO_DEV_NULL: GLib.SpawnFlags
SPAWN_STDOUT_TO_DEV_NULL: GLib.SpawnFlags
TYPE_BOOLEAN: GType
TYPE_BOXED: GType
TYPE_CHAR: GType
TYPE_DOUBLE: GType
TYPE_ENUM: GType
TYPE_FLAGS: GType
TYPE_FLAG_RESERVED_ID_BIT: builtins.int
TYPE_FLOAT: GType
TYPE_FUNDAMENTAL_MAX: builtins.int
TYPE_FUNDAMENTAL_SHIFT: builtins.int
TYPE_GSTRING: GType
TYPE_GTYPE: GType
TYPE_INT: GType
TYPE_INT64: GType
TYPE_INTERFACE: GType
TYPE_INVALID: GType
TYPE_LONG: GType
TYPE_NONE: GType
TYPE_OBJECT: GType
TYPE_PARAM: GType
TYPE_POINTER: GType
TYPE_PYOBJECT: GType
TYPE_RESERVED_BSE_FIRST: builtins.int
TYPE_RESERVED_BSE_LAST: builtins.int
TYPE_RESERVED_GLIB_FIRST: builtins.int
TYPE_RESERVED_GLIB_LAST: builtins.int
TYPE_RESERVED_USER_FIRST: builtins.int
TYPE_STRING: GType
TYPE_STRV: GType
TYPE_UCHAR: GType
TYPE_UINT: GType
TYPE_UINT64: GType
TYPE_ULONG: GType
TYPE_UNICHAR: GType
TYPE_VALUE: GType
TYPE_VARIANT: GType
VALUE_NOCOPY_CONTENTS: builtins.int
Warning: typing.Any
features: typing.Dict[str, bool]
glib_version: typing.Tuple[int, int, int]
pygobject_version: typing.Tuple[int, int, int]


GObject = Object



from unittest import TestCase

from blueman.main.DBusProxies import ProxyBase
from blueman.gobject import SingletonGObjectMeta


class TestDBusProxies(TestCase):
    def test_metaclass(self):
        self.assertIsInstance(ProxyBase, SingletonGObjectMeta)



from gettext import gettext as _
from typing import List, Union, Iterable, Dict, Optional, Callable, \
    TYPE_CHECKING, Tuple, Iterator, Mapping, Sequence

from gi.repository import GLib

from blueman.plugins.AppletPlugin import AppletPlugin
from operator import attrgetter

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    class SubmenuItemDict(TypedDict):
        text: str
        markup: bool
        icon_name: str
        sensitive: bool
        tooltip: Optional[str]
        callback: Callable[[], None]

    class MenuItemDict(SubmenuItemDict, total=False):
        submenu: Iterable["SubmenuItemDict"]


class MenuItem:
    def __init__(self, menu_plugin: "Menu", owner: AppletPlugin, priority: int, text: Optional[str], markup: bool,
                 icon_name: Optional[str], tooltip: Optional[str], callback: Optional[Callable[[], None]],
                 submenu_function: Optional[Callable[[], Iterable["SubmenuItemDict"]]], visible: bool, sensitive: bool):
        self._menu_plugin = menu_plugin
        self._owner = owner
        self._priority = priority
        self._text = text
        self._markup = markup
        self._icon_name = icon_name
        self._tooltip = tooltip
        self._callback = callback
        self._submenu_function = submenu_function
        self._visible = visible
        self._sensitive = sensitive

        assert text and icon_name and (callback or submenu_function) or \
            not any([text, icon_name, tooltip, callback, submenu_function])

    @property
    def owner(self) -> AppletPlugin:
        return self._owner

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def callback(self) -> Optional[Callable[[], None]]:
        return self._callback

    @property
    def visible(self) -> bool:
        return self._visible

    def _iter_base(self) -> Iterator[Tuple[str, Union[str, bool]]]:
        for key in ['text', 'markup', 'icon_name', 'tooltip', 'sensitive']:
            value = getattr(self, '_' + key)
            if value is not None:
                yield key, value

    def __iter__(self) -> Iterator[Tuple[str, Union[str, bool, List[Dict[str, Union[str, bool]]]]]]:
        yield from self._iter_base()
        submenu = list(self.submenu_items)
        if submenu:
            yield 'submenu', [dict(item) for item in submenu]

    @property
    def submenu_items(self) -> Iterable["SubmenuItem"]:
        if not self._submenu_function:
            return
        submenu_items = self._submenu_function()
        if not submenu_items:
            return
        for item in submenu_items:
            assert not set(item.keys()) - {'text', 'markup', 'icon_name', 'tooltip', 'callback', 'sensitive'}
            yield SubmenuItem(self._menu_plugin, self._owner, 0, item.get('text'), item.get('markup', False),
                              item.get('icon_name'), item.get('tooltip'), item.get('callback'), None, True,
                              item.get('sensitive', True))

    def set_text(self, text: str, markup: bool = False) -> None:
        self._text = text
        self._markup = markup
        self._menu_plugin.on_menu_changed()

    def set_icon_name(self, icon_name: str) -> None:
        self._icon_name = icon_name
        self._menu_plugin.on_menu_changed()

    def set_tooltip(self, tooltip: str) -> None:
        self._tooltip = tooltip
        self._menu_plugin.on_menu_changed()

    def set_visible(self, visible: bool) -> None:
        self._visible = visible
        self._menu_plugin.on_menu_changed()

    def set_sensitive(self, sensitive: bool) -> None:
        self._sensitive = sensitive
        self._menu_plugin.on_menu_changed()


class SubmenuItem(MenuItem):
    def __iter__(self) -> Iterator[Tuple[str, Union[str, bool]]]:
        yield from self._iter_base()


class Menu(AppletPlugin):
    __description__ = _("Provides a menu for the applet and an API for other plugins to manipulate it")
    __icon__ = "menu-editor"
    __author__ = "Walmis"
    __unloadable__ = False

    def on_load(self) -> None:
        self.__plugins_loaded = False

        self.__menuitems: List[MenuItem] = []

        self._add_dbus_signal("MenuChanged", "aa{sv}")
        self._add_dbus_method("GetMenu", (), "aa{sv}", self._get_menu)
        self._add_dbus_method("ActivateMenuItem", ("ai",), "", self._activate_menu_item)

    def __sort(self) -> None:
        self.__menuitems.sort(key=attrgetter('priority'))

    def add(self, owner: AppletPlugin, priority: int, text: Optional[str] = None, markup: bool = False,
            icon_name: Optional[str] = None, tooltip: Optional[str] = None,
            callback: Optional[Callable[[], None]] = None,
            submenu_function: Optional[Callable[[], Iterable["SubmenuItemDict"]]] = None,
            visible: bool = True, sensitive: bool = True) -> MenuItem:
        item = MenuItem(self, owner, priority, text, markup, icon_name, tooltip, callback, submenu_function, visible,
                        sensitive)
        self.__menuitems.append(item)
        if self.__plugins_loaded:
            self.__sort()
        self.on_menu_changed()
        return item

    def unregister(self, owner: AppletPlugin) -> None:
        for item in reversed(self.__menuitems):
            if item.owner == owner:
                self.__menuitems.remove(item)
        self.on_menu_changed()

    def on_plugins_loaded(self) -> None:
        self.__plugins_loaded = True
        self.__sort()

    def on_menu_changed(self) -> None:
        self._emit_dbus_signal("MenuChanged", self._get_menu())

    def _get_menu(self) -> List[Dict[str, GLib.Variant]]:
        return self._prepare_menu(dict(item) for item in self.__menuitems if item.visible)

    def _prepare_menu(self, data: Iterable[Mapping[str, Union[str, bool, Iterable[Mapping[str, Union[str, bool]]]]]]) \
            -> List[Dict[str, GLib.Variant]]:
        return [{k: self._build_variant(v) for k, v in item.items()} for item in data]

    def _build_variant(self, value: Union[str, bool, Iterable[Mapping[str, Union[str, bool]]]]) -> GLib.Variant:
        if isinstance(value, str):
            return GLib.Variant("s", value)
        if isinstance(value, bool):
            return GLib.Variant("b", value)
        return GLib.Variant("aa{sv}", self._prepare_menu(value))

    def _activate_menu_item(self, indexes: Sequence[int]) -> None:
        node = [item for item in self.__menuitems if item.visible][indexes[0]]
        for index in list(indexes)[1:]:
            node = [item for item in node.submenu_items if item.visible][index]
        if node.callback:
            node.callback()



from gettext import gettext as _
import logging
from typing import Optional, Any, List

from gi.repository import GLib
from blueman.plugins.AppletPlugin import AppletPlugin
from gettext import ngettext

from blueman.plugins.applet.StatusIcon import StatusIconProvider
from blueman.main.PluginManager import PluginManager


class ShowConnected(AppletPlugin, StatusIconProvider):
    __author__ = "Walmis"
    __depends__ = ["StatusIcon"]
    __icon__ = "blueman-active"
    __description__ = _("Adds an indication on the status icon when Bluetooth is active and shows the number of "
                        "connections in the tooltip.")

    def on_load(self) -> None:
        self.num_connections = 0
        self.active = False
        self.initialized = False
        self._handlers: List[int] = []
        self._handlers.append(self.parent.Plugins.connect('plugin-loaded', self._on_plugins_changed))
        self._handlers.append(self.parent.Plugins.connect('plugin-unloaded', self._on_plugins_changed))

    def on_unload(self) -> None:
        self.parent.Plugins.StatusIcon.set_text_line(1, None)
        self.num_connections = 0
        self.parent.Plugins.StatusIcon.icon_should_change()
        for handler in self._handlers:
            self.parent.Plugins.disconnect(handler)
        self._handlers = []

    def on_status_icon_query_icon(self) -> Optional[str]:
        if self.num_connections > 0:
            self.active = True
            return "blueman-active"
        else:
            self.active = False
            return None

    def enumerate_connections(self) -> bool:
        self.num_connections = 0
        for device in self.parent.Manager.get_devices():
            if device["Connected"]:
                self.num_connections += 1

        logging.info("Found %d existing connections" % self.num_connections)
        if (self.num_connections > 0 and not self.active) or \
                (self.num_connections == 0 and self.active):
            self.parent.Plugins.StatusIcon.icon_should_change()

        self.update_statusicon()

        return False

    def update_statusicon(self) -> None:
        if self.num_connections > 0:
            self.parent.Plugins.StatusIcon.set_text_line(0, _("Bluetooth Active"))
            self.parent.Plugins.StatusIcon.set_text_line(
                1,
                ngettext("<b>%(connections)d Active Connection</b>",
                         "<b>%(connections)d Active Connections</b>",
                         self.num_connections) % {"connections": self.num_connections})
        else:
            self.parent.Plugins.StatusIcon.set_text_line(1, None)
            if 'PowerManager' in self.parent.Plugins.get_loaded():
                status = self.parent.Plugins.PowerManager.get_bluetooth_status()
                if status:
                    self.parent.Plugins.StatusIcon.set_text_line(0, _("Bluetooth Enabled"))
                else:
                    self.parent.Plugins.StatusIcon.set_text_line(0, _("Bluetooth Disabled"))
            else:
                self.parent.Plugins.StatusIcon.set_text_line(0, "Blueman")

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            if not self.initialized:
                GLib.timeout_add(0, self.enumerate_connections)
                self.initialized = True
            else:
                GLib.timeout_add(1000, self.enumerate_connections)
        else:
            self.num_connections = 0
            self.update_statusicon()

    def on_device_property_changed(self, _path: str, key: str, value: Any) -> None:
        if key == "Connected":
            if value:
                self.num_connections += 1
            else:
                self.num_connections -= 1

            if (self.num_connections > 0 and not self.active) or (self.num_connections == 0 and self.active):
                self.parent.Plugins.StatusIcon.icon_should_change()

            self.update_statusicon()

    def on_adapter_added(self, _path: str) -> None:
        self.enumerate_connections()

    def on_adapter_removed(self, _path: str) -> None:
        self.enumerate_connections()

    def _on_plugins_changed(self, _pluginmngr: PluginManager, name: str) -> None:
        if name == "PowerManager":
            self.update_statusicon()



from typing import List

from gi.repository import GObject
from gi.repository import GLib
import socket
import subprocess
import logging
from blueman.Functions import have, get_local_interfaces
from blueman.bluemantyping import GSignals


class DhcpClient(GObject.GObject):
    __gsignals__: GSignals = {
        # arg: interface name eg. ppp0
        'connected': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
        'error-occurred': (GObject.SignalFlags.NO_HOOKS, None, (int,)),
    }

    COMMANDS = [
        ["dhclient", "-e", "IF_METRIC=100", "-1"],
        ["dhcpcd", "-m", "100"],
        ["udhcpc", "-t", "20", "-x", "hostname", socket.gethostname(), "-n", "-i"]
    ]

    quering: List[str] = []

    def __init__(self, interface: str, timeout: int = 30) -> None:
        """The interface name has to be trusted / sanitized!"""
        super().__init__()

        self._interface = interface
        self._timeout = timeout

        self._command = None
        for command in self.COMMANDS:
            path = have(command[0])
            if path:
                self._command = [path] + command[1:] + [self._interface]
                break

    def run(self) -> None:
        if not self._command:
            raise Exception("No DHCP client found, please install dhclient, dhcpcd, or udhcpc")

        if self._interface in DhcpClient.quering:
            raise Exception("DHCP already running on this interface")
        else:
            DhcpClient.quering.append(self._interface)

        self._client = subprocess.Popen(self._command)
        GLib.timeout_add(1000, self._check_client)
        GLib.timeout_add(self._timeout * 1000, self._on_timeout)

    def _on_timeout(self) -> bool:
        if not self._client.poll():
            logging.warning("Timeout reached, terminating DHCP client")
            self._client.terminate()
        return False

    def _check_client(self) -> bool:
        netifs = get_local_interfaces()
        status = self._client.poll()
        if status == 0:
            def complete() -> bool:
                ip = netifs[self._interface][0]
                logging.info(f"bound to {ip}")
                self.emit("connected", ip)
                return False

            GLib.timeout_add(1000, complete)
            DhcpClient.quering.remove(self._interface)
            return False
        elif status:
            logging.error(f"dhcp client failed with status code {status}")
            self.emit("error-occurred", status)
            DhcpClient.quering.remove(self._interface)
            return False
        else:
            return True



from typing import List, Dict, Optional
from typing_extensions import TypedDict

ERR: Dict[int, str]
RFCOMM_HANGUP_NOW: int
RFCOMM_RELEASE_ONHUP: int
RFCOMM_REUSE_DLC: int
RFCOMM_STATES: List[str]
RFCOMM_TTY_ATTACHED: int

class _RfcommDev(TypedDict):
    id: int
    channel: int
    flags: int
    state: str
    src: str
    dst: str

class _HciInfo(TypedDict):
    acl_mtu: int
    acl_pkts: int
    bdaddr: str
    dev_id: int
    features: List[int]
    flags: int
    link_mode: int
    link_policy: int
    name: str
    pkt_type: int
    sco_mtu: int
    stat: Dict[str, int]
    type: int

class BridgeException(Exception):
    errno: int
    def __init__(self, errno: int) -> None: ...

class ConnInfoReadError(Exception): ...

class RFCOMMError(Exception): ...

class conn_info:
    failed: bool
    def __init__(self, addr: str, hci_name: str) -> None: ...
    def deinit(self) -> None: ...
    def get_lq(self) -> int: ...
    def get_rssi(self) -> int: ...
    def get_tpl(self) -> int: ...
    def init(self) -> None: ...

def create_bridge(name: str = "pan1") -> None: ...
def create_rfcomm_device(local_address: str, remote_address: str, channel: int) -> int: ...
def destroy_bridge(name: str = "pan1") -> None: ...
def device_info(hci_name: str = "hci0") -> _HciInfo: ...
def page_timeout(hci_name: str = "hci0") -> float: ...
def get_rfcomm_channel(uuid: int, bdaddr: str) -> Optional[int]: ...
def release_rfcomm_device(id: int) -> int: ...
def rfcomm_list() -> List[_RfcommDev]: ...



from blueman.services.DialupNetwork import DialupNetwork as DialupNetwork
from blueman.services.GroupNetwork import GroupNetwork as GroupNetwork
from blueman.services.NetworkAccessPoint import NetworkAccessPoint as NetworkAccessPoint
from blueman.services.SerialPort import SerialPort as SerialPort
from blueman.services.Functions import get_service as get_service, get_services as get_services

__all__ = ["DialupNetwork", "GroupNetwork", "NetworkAccessPoint", "SerialPort", "get_service", "get_services"]



from enum import Enum
from gettext import gettext as _
import logging
from typing import Callable, Any, Optional

from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluez.Adapter import Adapter

from gi.repository import GLib

from blueman.plugins.applet.StatusIcon import StatusIconProvider


class PowerStateListener:
    def on_power_state_changed(self, manager: "PowerManager", state: bool) -> None:
        ...


class PowerStateHandler:
    def on_power_state_query(self) -> "PowerManager.State":
        ...

    def on_power_state_change_requested(self, manager: "PowerManager", state: bool,
                                        cb: Callable[[bool], None]) -> None:
        ...


class PowerManager(AppletPlugin, StatusIconProvider):
    __depends__ = ["StatusIcon", "Menu"]
    __unloadable__ = True
    __description__ = _("Controls Bluetooth adapter power states")
    __author__ = "Walmis"
    __icon__ = "gnome-power-manager"

    __gsettings__ = {
        "schema": "org.blueman.plugins.powermanager",
        "path": None
    }

    __options__ = {
        "auto-power-on": {
            "type": bool,
            "default": True,
            "name": _("Auto power-on"),
            "desc": _("Automatically power on adapters")
        }
    }

    class State(Enum):
        ON = 2
        OFF = 1
        OFF_FORCED = 0

    def on_load(self) -> None:
        self.item = self.parent.Plugins.Menu.add(self, 0, text=_("<b>Turn Bluetooth _Off</b>"), markup=True,
                                                 icon_name="blueman-disabled", tooltip=_("Turn off all adapters"),
                                                 callback=self.on_bluetooth_toggled)
        self.adapter_state = True
        self.current_state = True

        self.request_in_progress = False

        self._add_dbus_signal("BluetoothStatusChanged", "b")
        self._add_dbus_method("SetBluetoothStatus", ("b",), "", self.request_power_state)
        self._add_dbus_method("GetBluetoothStatus", (), "b", self.get_bluetooth_status)

    def on_unload(self) -> None:
        self.parent.Plugins.Menu.unregister(self)

    @property
    def CurrentState(self) -> bool:
        return self.current_state

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            def timeout() -> bool:
                self.adapter_state = self.get_adapter_state()
                if self.get_option("auto-power-on"):
                    self.request_power_state(True, force=True)
                else:
                    self.request_power_state(self.adapter_state)
                return False

            GLib.timeout_add(1000, timeout)

    def get_adapter_state(self) -> bool:
        adapters = self.parent.Manager.get_adapters()
        for adapter in adapters:
            if not adapter["Powered"]:
                return False
        return bool(adapters)

    def set_adapter_state(self, state: bool) -> None:
        try:
            logging.info(state)
            adapters = self.parent.Manager.get_adapters()
            for adapter in adapters:
                adapter.set("Powered", state)

            self.adapter_state = state
        except Exception:
            logging.error("Exception occurred", exc_info=True)

    class Callback:
        def __init__(self, parent: "PowerManager", state: bool):
            self.parent = parent
            self.num_cb = 0
            self.called = 0
            self.state = state
            self.success = False
            self.timer = GLib.timeout_add(5000, self.timeout)

        def __call__(self, result: bool) -> None:
            self.called += 1

            if result:
                self.success = True

            self.check()

        def check(self) -> None:
            if self.called == self.num_cb:
                GLib.source_remove(self.timer)
                logging.info("callbacks done")
                self.parent.set_adapter_state(self.state)
                self.parent.update_power_state()
                self.parent.request_in_progress = False

        def timeout(self) -> bool:
            logging.info("Timeout reached while setting power state")
            self.parent.update_power_state()
            self.parent.request_in_progress = False
            return False

    def request_power_state(self, state: bool, force: bool = False) -> None:
        if self.current_state != state or force:
            if not self.request_in_progress:
                self.request_in_progress = True
                logging.info(f"Requesting {state}")
                cb = PowerManager.Callback(self, state)

                handlers = list(self.parent.Plugins.get_loaded_plugins(PowerStateHandler))
                for handler in handlers:
                    handler.on_power_state_change_requested(self, state, cb)
                cb.num_cb = len(handlers)
                cb.check()
            else:
                logging.info("Another request in progress")

    # queries other plugins to determine the current power state
    def update_power_state(self) -> None:
        rets = [plugin.on_power_state_query()
                for plugin in self.parent.Plugins.get_loaded_plugins(PowerStateHandler)]

        off = any(x != self.State.ON for x in rets) or not self.adapter_state
        foff = self.State.OFF_FORCED in rets
        on = self.State.ON in rets or self.adapter_state

        new_state = True
        if foff or off:

            self.item.set_text(_("<b>Turn Bluetooth _On</b>"), markup=True)
            self.item.set_icon_name("blueman")
            self.item.set_tooltip(_("Turn on all adapters"))
            self.item.set_sensitive(not foff)

            new_state = False

        elif on and not self.current_state:

            self.item.set_text(_("<b>Turn Bluetooth _Off</b>"), markup=True)
            self.item.set_icon_name("blueman-disabled")
            self.item.set_tooltip(_("Turn off all adapters"))
            self.item.set_sensitive(True)

            new_state = True

        logging.info(f"off {off} | foff {foff} | on {on} | current state {self.current_state} | new state {new_state}")

        if self.current_state != new_state:
            logging.info(f"Signalling {new_state}")
            self.current_state = new_state

            self._emit_dbus_signal("BluetoothStatusChanged", new_state)
            for plugin in self.parent.Plugins.get_loaded_plugins(PowerStateListener):
                plugin.on_power_state_changed(self, new_state)

            if new_state:
                self.parent.Plugins.StatusIcon.set_text_line(0, _("Bluetooth Enabled"))
                self.parent.Plugins.StatusIcon.query_visibility(delay_hiding=True)
            else:
                self.parent.Plugins.StatusIcon.set_text_line(0, _("Bluetooth Disabled"))
                self.parent.Plugins.StatusIcon.query_visibility()
            self.parent.Plugins.StatusIcon.icon_should_change()

    def get_bluetooth_status(self) -> bool:
        return self.current_state

    def on_adapter_property_changed(self, _path: str, key: str, value: Any) -> None:
        if key == "Powered":
            if value and not self.current_state:
                logging.warning("adapter powered on while in off state, turning bluetooth on")
                self.request_power_state(True)

            self.adapter_state = self.get_adapter_state()
            self.update_power_state()

    def on_bluetooth_toggled(self) -> None:
        self.request_power_state(not self.current_state)

    def on_status_icon_query_icon(self) -> Optional[str]:
        return "blueman-disabled" if not self.get_bluetooth_status() else None

    def on_adapter_added(self, path: str) -> None:
        adapter = Adapter(obj_path=path)
        adapter.set("Powered", self.adapter_state)
